/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dashboard.
 *
 * The Initial Developer of the Original Code is
 *   Dave Raggett.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

// sqlite database for data track
// together with common queries
dashboard_overlay.register_database({
  db: null,  // handle to dashboard database
  permissions: null, // handle to permissions database

  start_up: function (context)
  {
    this.context = context;

    // ensure that only one database connection exists for all windows
    if (context.window_count == 0)
    {
      // open handle to dashboard database
      var file = Components.classes["@mozilla.org/file/directory_service;1"]  
                     .getService(Components.interfaces.nsIProperties)  
                     .get("ProfD", Components.interfaces.nsIFile);

      file.append("dashboard.sqlite");  
  
      var storage = Components.classes["@mozilla.org/storage/service;1"]  
                               .getService(Components.interfaces.mozIStorageService);

      // Creates the file if it does not exist
      context.dashboard_db = storage.openDatabase(file);

      // also open handle for Firefox's permissions database
      var file2 = Components.classes["@mozilla.org/file/directory_service;1"]  
                     .getService(Components.interfaces.nsIProperties)  
                     .get("ProfD", Components.interfaces.nsIFile);

      file2.append("permissions.sqlite");  

      context.dashboard_permissions = storage.openDatabase(file2);

      // and open handle for Firefox's DOM storage database
      var file3 = Components.classes["@mozilla.org/file/directory_service;1"]  
                     .getService(Components.interfaces.nsIProperties)  
                     .get("ProfD", Components.interfaces.nsIFile);

      file3.append("webappsstore.sqlite");  

      context.dashboard_dom_storage = storage.openDatabase(file3);

      if (!context.dashboard_dom_storage)
        alert("couldn't open dom storage db");

      // create data track tables if they don't exist
      this.create_data_track(context.dashboard_db);
    }

    this.db = context.dashboard_db;
    this.permissions = context.dashboard_permissions;
    this.dom_storage = context.dashboard_dom_storage;
  },

  shut_down: function (context)
  {
    // MUST first call finalize() on all remaining
    // mozIStorageStatement objects.
    if (context.window_count == 0)
    {
      try
      {
        this.db.close();
        this.db = null;

        this.permissions.close();
        this.permissions = null;

        this.dom_storage.close();
        this.dom_storage = null;
     }
     catch (e) { };
    }
  },

  reset: function (all)
  {
    this.drop_data_track(this.db, all);
    this.create_data_track(this.db);
  },

  // poll Firefox private browsing mode
  // we use this to suppress logging
  private_mode: function ()
  {
    var pbs = Components.classes["@mozilla.org/privatebrowsing;1"]  
               .getService(Components.interfaces.nsIPrivateBrowsingService);  
    return pbs.privateBrowsingEnabled; 
  },

  sync_command: function (sql, sql2)
  {
    var statement = this.db.createStatement(sql);
    var transaction = false;

    if (this.db.transactionInProgress)
    {
      transaction = true;
      this.db.beginTransactionAs(this.db.TRANSACTION_DEFERRED);
    }

    try 
    {
      try
      {
        statement.execute();
      }
      finally
      {
        statement.reset();

        if (statement.finalize)
          statement.finalize();
      }
    }
    catch (e)
    { 
      if (sql2)
        dashboard_overlay.database.sync_command(sql2);
    }

    if (transaction)
      this.db.commitTransaction();
  },

  // database command that doesn't involve generating
  // results, uses a synchronous method for Firefox < 3.5
  // and an asynch method for Firefox 3.5 and later
  command: function (sql, sql2)
  {
    try {
    var statement = this.db.createStatement(sql);
    } catch (e) { alert(sql + "\n\n" + e); return; }

    if (typeof statement.executeAsync != "undefined")
    {

      statement.executeAsync({
        // no results, so this is a no-op
        handleResult: function(aResultSet) 
        {
        },

        handleError: function(aError)
        {
          // alert("Error: " + aError.message + "\n"+sql);
        },

        // called after last row has been passed to handleResult
        handleCompletion: function(aReason)
        {
          if (aReason != Components.interfaces.mozIStorageStatementCallback.REASON_FINISHED)
          {
            if (sql2)
              dashboard_overlay.database.command(sql2);
            else
              alert("Query canceled or aborted! " + sql +
                    ", " + dashboard_overlay.database.db.lastErrorString);
          }
        }
      });
    }
    else // synchronous execution of sql
    {
      var transaction = false;

      if (this.db.transactionInProgress)
      {
        transaction = true;
        this.db.beginTransactionAs(this.db.TRANSACTION_DEFERRED);
      }

      try 
      {
        try
        {
          statement.execute();
        }
        finally
        {
          statement.reset();

          if (statement.finalize)
            statement.finalize();
        }
      }
      catch (e)
      { 
        if (sql2)
          dashboard_overlay.database.command(sql2);
      }

      if (transaction)
        this.db.commitTransaction();
    }
  },

  // database SELECT command that generates results
  // this uses a synchronous method for Firefox < 3.5
  // use asynchronous method for later versions
  select: function (sql, db, handler)
    {
      let i, k, statement, transaction = false;

      if (this.db == null)
        this.start_up();

      if (!db)
        db = this.db;

      try 
      {
        statement = db.createStatement(sql);
      }
      catch (e)
      {
        if (db != this.dom_storage)
          alert("Couldn't execute: " + sql);
        return;
      }

      if (typeof statement.executeAsync != "undefined")
      {
        statement.executeAsync({
          // called multiple times with one or more rows each time
          handleResult: function(aResultSet)
          {
            for (let row = aResultSet.getNextRow(); row; row = aResultSet.getNextRow())
              handler.process_row(row);
          },

          handleError: function(aError) 
          {
            statement.reset();

            if (statement.finalize)
              statement.finalize();

            alert("Error: " + aError.message);
          },

          // called after last row has been passed to handleResult
          handleCompletion: function(aReason)
          {
            statement.reset();

            if (statement.finalize)
              statement.finalize();

            handler.finalize();

            if (aReason != Components.interfaces.mozIStorageStatementCallback.REASON_FINISHED)
              alert("Query canceled or aborted! " + 
                         dashboard_overlay.database.db.lastErrorString);
          }
        });
      }
      else // use synchronous query
      {
        if (db.transactionInProgress)
        {
          transaction = true;
          db.beginTransactionAs(db.TRANSACTION_DEFERRED);
        }

        try
        {
          while (statement.executeStep())
          {
            var row = { "data": {} };

            // mimic asynchronous interface for column values
            row.getResultByName = function (name)
            {
              return row.datum[name];
            }

            for (i = 0, k = statement.columnCount; i < k; ++i)
              row.data[statement.getColumnName(i)] = statement.getUTF8String(i);

            handler.process_row(row);
          }
        }
        finally
        {
          statement.reset();
          statement.close();
        }

        if (transaction)
          db.commitTransaction();

        handler.finalize();
      }
    },

  // used to create the sqlite tables for the data track
  // this is done synchronously as it should be fast given
  // that the tables will normally already exist
  create_data_track: function (db)
  {
    // direct resource to resource relation
    db.executeSimpleSQL(
      "CREATE TABLE IF NOT EXISTS relations (" +
          "parent TEXT," +
          "child TEXT," +
          "offsite INTEGER," +
          "time DATE," +
          "PRIMARY KEY (parent, child)" +
      ")");

    db.executeSimpleSQL(
      "CREATE INDEX IF NOT EXISTS relations_parent_index ON relations (parent)");

    // could replace this by DEFAULT CURRENT_TIMESTAMP on time column
    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS insert_relations_time AFTER" +
      " INSERT ON relations\n" +
      "BEGIN\n" +
      " UPDATE relations SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    // needed since DEFAULT only effects INSERT not UPDATE
    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS update_relations_time AFTER" +
      " UPDATE ON relations\n" +
      "BEGIN\n" +
      " UPDATE relations SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    // page to direct and indirect 3rd parties
    db.executeSimpleSQL(
      "CREATE TABLE IF NOT EXISTS parties (" +
          "page_host TEXT," +
          "third_party TEXT," +
          "offsite INTEGER," +
          "time DATE," +
          "PRIMARY KEY (page_host, third_party)" +
      ")");

    db.executeSimpleSQL(
      "CREATE INDEX IF NOT EXISTS parties_page_host_index ON parties (page_host)");

    db.executeSimpleSQL(
      "CREATE INDEX IF NOT EXISTS parties_third_party_index ON parties (third_party)");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS insert_parties_time AFTER" +
      " INSERT ON parties\n" +
      "BEGIN\n" +
      " UPDATE parties SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS update_parties_time AFTER" +
      " UPDATE ON parties\n" +
      "BEGIN\n" +
      " UPDATE parties SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    // tracks data POSTed via HTTP
    // redundant copies of domain and date for efficient
    // queries on who you gave what data to and when
    db.executeSimpleSQL(
      "CREATE TABLE IF NOT EXISTS http_data (" +
          "name TEXT," +
          "value TEXT," +
          "host TEXT," +
          "posted INTEGER," +
          "form INTEGER," +
          "time TEXT" +
      ")");

    db.executeSimpleSQL(
      "CREATE INDEX IF NOT EXISTS http_data_host_index ON http_data (host)");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS insert_http_data_time AFTER" +
      " INSERT ON http_data\n" +
      "BEGIN\n" +
      " UPDATE http_data SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS update_http_data_time AFTER" +
      " UPDATE ON http_data\n" +
      "BEGIN\n" +
      " UPDATE http_data SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    db.executeSimpleSQL(
      "CREATE TABLE IF NOT EXISTS site_info (" +
          "host TEXT PRIMARY KEY," +
          "visited INTEGER," +
          "prefs INTEGER," +
          "session_cookies INTEGER," +
          "lasting_cookies INTEGER," +
          "flash_cookies INTEGER," +
          "int_3rd_parties INTEGER," +
          "ext_3rd_parties INTEGER," +
          "int_3rd_party_session_cookies INTEGER," +
          "int_3rd_party_lasting_cookies INTEGER," +
          "int_3rd_party_flash_cookies INTEGER," +
          "ext_3rd_party_session_cookies INTEGER," +
          "ext_3rd_party_lasting_cookies INTEGER," +
          "ext_3rd_party_flash_cookies INTEGER," +
          "dom_storage INTEGER," +
          "html5_pings INTEGER," +
          "invisible_images INTEGER," +
          "suspicious_urls INTEGER," +
          "geo_permission INTEGER," +
          "p3p INTEGER," +
          "time DATE" +
      ")");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS insert_site_info_time AFTER" +
      " INSERT ON site_info\n" +
      "BEGIN\n" +
      " UPDATE parties SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");

    db.executeSimpleSQL(
      "CREATE TRIGGER IF NOT EXISTS update_site_info_time AFTER" +
      " UPDATE OF visited ON site_info\n" +
      "BEGIN\n" +
      " UPDATE site_info SET time = DATETIME('NOW')\n" + 
      " WHERE rowid = new.rowid;\n" +
      "END");
  },

  drop_data_track: function (db, all)
  {
    db.executeSimpleSQL("DROP TABLE IF EXISTS http_data");
    db.executeSimpleSQL("DROP TABLE IF EXISTS parties");
    db.executeSimpleSQL("DROP TABLE IF EXISTS relations");

    // keep site_info to retain user's site preferences?

    if (all)
      db.executeSimpleSQL("DROP TABLE IF EXISTS site_info");
  },

  iso8061: function (date)
  {
     if (!date)
       date = new Date();

     return date.getFullYear() + "-" + (1 + date.getMonth()) + "-" + date.getDate()
       + " " + date.getHours() + ":" + date.getMinutes() + ":" +date.getSeconds();
  },

  log_3rd_parties: function (status)
  {
    let private_mode = this.private_mode();
    let count = 0;
    let page_host = status.page_host;
    let page_domain = status.page_domain;
    let parties = status.parties;

    if (!private_mode)
    {
      for (party in parties)
      {
        ++count;
        let sql = "INSERT OR REPLACE INTO parties " +
                  "(page_host, third_party, offsite) " +
                  "VALUES(" +
                  this.quote(page_host) + "," +
                  this.quote("" + party) + "," +
                  parties[party] +
                  ")";

        this.command(sql);
      }
    }

    status.parties = {}; // reclaim memory
  },

  // called by http observer on page load or timer
  // may want to switch to execAsynch on statement
  // which allows you to set error handler+continuation
  // doc.baseURIObject is page's nsURI object and
  // is used to identify 3rd party requests
  // note this has been greatly trimmed down and
  // could easily be expanded to log more details
  log_http: function (requests)
  {
    let private_mode = this.private_mode();
    var i, j, request, sql, param;

    for (i = 0; i < requests.length; ++i)
    {
      request = requests[i];

      // want to reduce redundant records
      if (request.ref_host && request.host != request.ref_host)
      {
        sql = "INSERT OR REPLACE INTO relations " +
              "(parent, child, offsite) " +
              "VALUES(" +
              this.quote(request.ref_host) + "," +
              this.quote(request.host) + "," +
              request.offsite +
              ")";

        if (!private_mode && request.page_host)
          this.command(sql);
      }

      for (j = 0; j < request.query_params.length; ++j)
      {
        param = request.query_params[j];

        sql = "INSERT INTO http_data " +
              "(name, value, host, posted, form) " +
              "VALUES(" +
              this.quote(param.name) + "," +
              this.quote(param.value) + "," +
              this.quote(request.host) + "," +
              0 + "," +      // from post?
              param.form +   // from form?
              ")";

        if (!private_mode)
          this.command(sql);
      }

      for (j = 0; j < request.post_params.length; ++j)
      {
        param = request.post_params[j];

        if (param)
        {
          sql = "INSERT INTO http_data " +
              "(name, value, host, posted, form) " +
              "VALUES(" +
              this.quote(param.name) + "," +
              this.quote(param.value) + "," +
              this.quote(request.host) + "," +
              1 + "," +      // from post?
              param.form +   // from form?
              ")";

          if (!private_mode)
            this.command(sql);
        }
      }
    }
  },

  // update site info for later sharing with others
  // omits p3p which is saved asynchronously due to
  // time involved in searching web for the policy
  // sets feature counts to max(old, new)
  update_site_info: function (host, status)
  {
    var sql = "SELECT " +
       "session_cookies,lasting_cookies,flash_cookies," +
       "int_3rd_parties,ext_3rd_parties," +
       "int_3rd_party_session_cookies,int_3rd_party_lasting_cookies," +
       "int_3rd_party_flash_cookies,ext_3rd_party_session_cookies," +
       "ext_3rd_party_lasting_cookies,ext_3rd_party_flash_cookies," +
       "dom_storage,html5_pings,invisible_images," +
       "suspicious_urls,geo_permission" + " " +
       "FROM site_info WHERE host=" + this.quote(host);

    let found = false;
    let database = this;

    let handler = 
    {
      set_max: function (row, col)
      {
        let n = row.getResultByName(col);
        let m = status[col];

        if (n > m)
          status[col] = n;
      },

      process_row: function (row)
      {
        let set = this.set_max;

        set(row, "session_cookies");
        set(row, "lasting_cookies");
        set(row, "flash_cookies");
        set(row, "int_3rd_parties");
        set(row, "ext_3rd_parties");
        set(row, "int_3rd_party_session_cookies");
        set(row, "int_3rd_party_lasting_cookies");
        set(row, "int_3rd_party_flash_cookies");
        set(row, "ext_3rd_party_session_cookies");
        set(row, "ext_3rd_party_lasting_cookies");
        set(row, "ext_3rd_party_flash_cookies");
        set(row, "dom_storage");
        set(row, "html5_pings");
        set(row, "invisible_images");
        set(row, "suspicious_urls");
        set(row, "geo_permission");

        let sql = "UPDATE site_info SET " +
          "session_cookies=" + status.session_cookies + "," +
          "lasting_cookies=" + status.lasting_cookies + "," +
          "flash_cookies=" + status.flash_cookies + "," +
          "int_3rd_parties=" + status.int_3rd_parties + "," +
          "ext_3rd_parties=" + status.ext_3rd_parties + "," +
          "int_3rd_party_session_cookies=" + status.int_3rd_party_session_cookies + "," +
          "int_3rd_party_lasting_cookies=" + status.int_3rd_party_lasting_cookies + "," +
          "int_3rd_party_flash_cookies=" + status.int_3rd_party_flash_cookies + "," +
          "ext_3rd_party_session_cookies=" + status.ext_3rd_party_session_cookies + "," +
          "ext_3rd_party_lasting_cookies=" + status.ext_3rd_party_lasting_cookies + "," +
          "ext_3rd_party_flash_cookies=" + status.ext_3rd_party_flash_cookies + "," +
          "dom_storage=" + status.dom_storage + "," +
          "html5_pings=" + status.html5_pings + "," +
          "invisible_images=" + status.invisible_images + "," +
          "suspicious_urls=" + status.suspicious_urls + "," +
          "geo_permission= " + status.geo_permission + " " +
          "WHERE host=" + database.quote(host);

        database.command(sql);
        found = true;
      },

      // called after all rows of query have been delivered
      finalize: function ()
      {
        if (!found)
        {
          let sql = "INSERT INTO site_info (" +
            "host,visited,prefs,session_cookies,lasting_cookies," +
            "flash_cookies,int_3rd_parties,ext_3rd_parties," +
            "int_3rd_party_session_cookies,int_3rd_party_lasting_cookies," +
            "int_3rd_party_flash_cookies,ext_3rd_party_session_cookies," +
            "ext_3rd_party_lasting_cookies,ext_3rd_party_flash_cookies," +
            "dom_storage, html5_pings,invisible_images," +
            "suspicious_urls,geo_permission" + 
          ") VALUES(" +
            database.quote(host) + "," +
            1+ "," +
            status.prefs + "," +
            status.session_cookies + "," +
            status.lasting_cookies + "," +
            status.flash_cookies + "," +
            status.int_3rd_parties + "," +
            status.ext_3rd_parties + "," +
            status.int_3rd_party_session_cookies + "," +
            status.int_3rd_party_lasting_cookies + "," +
            status.int_3rd_party_flash_cookies + "," +
            status.ext_3rd_party_session_cookies + "," +
            status.ext_3rd_party_lasting_cookies + "," +
            status.ext_3rd_party_flash_cookies + "," +
            status.dom_storage + "," +
            status.html5_pings + "," +
            status.invisible_images + "," +
            status.suspicious_urls + "," +
            status.geo_permission +
          ")";

          database.command(sql);
        }
      }
    };

    database.select(sql, this.db, handler);
  },

  // update or insert record for given host
  // that only changes the given column name
  // sqlite lacks insert or update on duplicate key
  update_or_insert: function (host, property, value, sync)
  {
      // fails if record already present for this host
      var sql1 = "INSERT INTO site_info (host,"+
                property + ") VALUES(" +
                this.quote(host) + "," +
                value + 
                ")";

      // whereupon we execute the update statement
      var sql2 = "UPDATE site_info SET " +
                property + "=" + value +
                " WHERE host=" + this.quote(host);

      if (!this.private_mode())
      {
        if (sync)
        {
          this.sync_command(sql1, sql2);
        }
        else
        {
          this.command(sql1, sql2);
        }
      }
  },

  p3p_log: function (host, req)
  {
    var p3p = (req && req.responseXML) ? 1 : 0;

    if (!this.private_mode())
      this.update_or_insert(host, "p3p", p3p);
  },

/*
  // find all session cookies with given name
  // and check for match on http server host
  // e.g. .lovefilm.com cf www.lovefilm.com
  session_cookie: function (host, name) {
    var host_len = host.length;

    var sql = "SELECT host FROM cookies WHERE name = '" +
              name + "' AND session <> 0";

    var dataset = this.sync_get(sql, this.db);

    for (var i = 0; dataset && i < dataset.length; ++i)
    {
      pattern = dataset[0]["host"];
      var len = pattern.length;

      if (len <= hostlen)
      {
        var hs = host.substring(hostlen - len);

        if (hs == pattern)
          return true;
      }
    }

    return false;
  },

  // find all persistent cookies with given name
  // and check for match on http server host
  // e.g. .lovefilm.com cf www.lovefilm.com
  persistent_cookie: function (host, name) {
    var host_len = host.length;

    var sql = "SELECT host FROM cookies WHERE name = '" +
              name + "' AND session = 0";

    var dataset = this.sync_get(sql, this.db);

    for (var i = 0; dataset && i < dataset.length; ++i)
    {
      pattern = dataset[0]["host"];
      var len = pattern.length;

      if (len <= hostlen)
      {
        var hs = host.substring(hostlen - len);

        if (hs == pattern)
          return true;
      }
    }

    return false;
  },
*/
  // update site preferences where prefs is an integer value
  // encoded as a bit vector
  set_prefs: function (host, prefs)
  {
    console.log("database:set_prefs " + prefs);
    this.update_or_insert(host, "prefs", prefs);

    // and synchronise session cache
    dashboard_overlay.sync_prefs(host, prefs);
  },

  // synchronous query returning bit vector of site preferences
  get_prefs: function (host, db)
  {
    var prefs;

    let sql = "SELECT prefs FROM site_info WHERE host = '" +
              host + "'";

    let dataset = this.sync_get(sql, db);

    if (dataset && dataset.length > 0)
      prefs = dataset[0]["prefs"];

    return prefs;
  },

  // fast synchronous read and write
  mark_visited: function (browser)
  {
    let status = browser.dashboard_status;
    let visited = 0;

    try
    {
      let host = status.page_host;

      if (host)
      {
        var sql = "SELECT visited FROM site_info WHERE host = '" +
                  host + "'";

        var dataset = this.sync_get(sql, this.db);

        if (dataset && dataset.length > 0)
          visited = dataset[0]["visited"];

       this.update_or_insert(status.page_host, "visited", ++visited, true);
      }
    }
    catch (e) { }

    status.visited = visited;
  },

  extract_host: function (referrer)
  {
    var page_host = null;
    var i;

    if (referrer && (i = referrer.indexOf("://")) > 0)
    {
      page_host = referrer.substr(i+3);

      i = page_host.indexOf("/");
      page_host = page_host.substr(0, i);
    }

    return page_host;
  },

  sync_get: function (sql, db)
  {
    var dataset;
    var transaction = false;
    var statement = db.createStatement(sql);  // throws exception on error

    if (db.transactionInProgress)
    {
      transaction = true;
      db.beginTransactionAs(db.TRANSACTION_DEFERRED);
    }

    try
    {
      var dataset = [];

     while (statement.executeStep())
     {
       var row = [];

       for (var i = 0, k = statement.columnCount; i < k; ++i)
         row[statement.getColumnName(i)] = statement.getUTF8String(i);

       dataset.push(row);
     }
    }
    finally
    {
      statement.reset();
      //statement.close();
    }

    if (transaction)
      db.commitTransaction();

    return dataset;
  },

  quote: function (s)
  {
     if (s)
     {
       s = s.replace(/\'/g, "''");
       return "'" + s + "'";
     }

     return "null";
  },

  bool: function (v)
  {
     return (v ? "1" : "0");
  }
});

