/*
 *      graphviz.c -- generate graphviz dot file from dashboard data
 *
 *      Copyright (c) 2011 Dave Raggett
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>   // for strlen
#include <strings.h>
#include <sqlite3.h>

#include "hashtable.h"
#include "analyst.h"
#include "top_sites.h"

#define GRAPHVIZ_FILE "dashboard.dot"

static struct hashtable *sites = NULL;

static void print_comment_open(FILE *fp)
{
  fprintf(fp, "\n/*\n");
}

static void print_comment_line(FILE *fp, const char *format, ...)
{
  va_list args;
  va_start(args, format);

  fputs(" * ", fp);
  vfprintf(fp, format, args);
  putc('\n', fp);

  va_end(args);
}

static void print_comment_close(FILE *fp)
{
  fputs(" */\n\n", fp);
}

static void print_diagraph_open(FILE *fp, const char *format, ...)
{
  va_list args;
  va_start(args, format);

  fputs("digraph ", fp);
  vfprintf(fp, format, args);
  fputs(" {\n", fp);

  va_end(args);
}

static void print_diagraph_arc(FILE *fp, char *from, char *to, int flag)
{
  if (flag)
    fprintf(fp, "  %s -> %s;\n", from, to);
  else
    fprintf(fp, "  %s -> %s [style=dotted];\n", from, to);
}

static void print_diagraph_close(FILE *fp)
{
  fputs("}\n", fp);
}

static void declare_node(FILE *fp, char *host, bool visible)
{
  if (visible)
    fprintf(fp, "  %s [shape = ellipse]\n", host);
  else
    fprintf(fp, "  %s [shape = box]\n", host);
}

static char *escape(char *domain)
{
  int c;
  char *str = (char *)malloc(strlen(domain)+3);
  assert(str);

  char *p = str, *q = domain;

  *p++ = 's';
  *p++ = '_';

  while ((c = *q++))
  {
    *p++ = (c == '.'  || c == '-') ? '_' : c;
  }

  *p = '\0';
  return str;
}

static char *get_node(FILE *fp, char *host)
{
  char *id = (char *)hashtable_search(sites, host);

  if (id == null)
  {
    id = escape(host);
    host = strdup(host);
    assert(host);
    assert(hashtable_insert(sites, host, id, 1));

    fprintf(fp, "  %s [shape=\"ellipse\",label=\"%s\" color=black]\n",
            id, host);
  }

  return id;
}


// 1st arg is context passed in 4th arg of sqlite3_exec
// 2nd arg is number of columns
// 3rd arg is an array of pointers to column values
// 4th arg is an array of pointers to column names
// note: column value may be null so check for null pointer
static int arcs(void *fp, int argc, char **argv, char **azColName)
{
  char *page_host = argv[0];
  char *third_party = argv[1];
  char *offsite = argv[2];

  char *host_id = get_node((FILE *)fp, page_host);
  char *third_pty_id = get_node((FILE *)fp, third_party);

  print_diagraph_arc((FILE *)fp, host_id, third_pty_id, strcmp(offsite, "1"));

  return 0;
}


static void third_parties(FILE *fp, sqlite3 *db, char *host)
{
  char sql[1024];
  char *zErrMsg = null;

  sprintf(sql,
      "SELECT page_host,third_party,offsite FROM parties WHERE page_host = \"www.%s\"",
      host);

  int rc = sqlite3_exec(db, sql, arcs, fp, &zErrMsg);

  if( rc != SQLITE_OK )
  {
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
}

// use direct parent-child relationships
static void graph_direct(FILE *fp, sqlite3 *db)
{
  char *sql = "SELECT parent,child,offsite FROM relations";
  char *zErrMsg = null;

  int rc = sqlite3_exec(db, sql, arcs, fp, &zErrMsg);

  if( rc != SQLITE_OK )
  {
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
}

// consider all website links to third parties direct and indirect
static void graph_sites(FILE *fp, sqlite3 *db, int start, int stop)
{
  int i;
  SiteDesc *desc;

  if (start < 1 || stop > NUM_SITES)
  {
    fprintf(stderr, "Out of range (%d  to %d)\n", start, stop);
    sqlite3_close(db);
    exit(1);
  }

  for (i = start - 1; i < stop; ++i)
  {
    desc = top_sites + i;
    get_node(fp, desc->domain);
    third_parties(fp, db, desc->domain);
  }
}

static int compare_strings(void *key1, void *key2)
{
  return !strcmp((char *)key1, (char *)key2);
}

void graph(sqlite3 *db)
{
  sites = create_hashtable(8192, oat_hash2, compare_strings);

  if (!sites)
  {
    fprintf(stderr, "Can't create hash table\n");
    sqlite3_close(db);
    exit(1);
  }

  char *name = GRAPHVIZ_FILE;
  FILE *fp = fopen(name, "w+");

  if (!fp)
  {
    fprintf(stderr, "Can't open output file %s\n", name);
    sqlite3_close(db);
    exit(1);
  }

  print_comment_open(fp);
  //print_comment_line(fp, "@command =  neato");
  //print_comment_line(fp, "");
  print_comment_line(fp, "graphiz dot file generated from privacy dashboard data");
  print_comment_close(fp);

  fprintf(fp, "digraph third_parties {\ngraph [splines=true,overlap=true]\n");

  graph_direct(fp, db);
  //graph_sites(fp, db, 1, 1000);
  print_diagraph_close(fp);

  fclose(fp);
  fprintf(stderr, "written graph to %s\n", name);
  hashtable_destroy(sites, 1);
}

